package gui;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import gui.model.Engine;
import gui.model.Switch;
import gui.model.Settings;
import gui.view.EngineController;
import gui.view.TurntableController;
import gui.view.RootLayoutController;
import gui.view.SettingsController;
import gui.view.SwitchListController;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.Tab;
import javafx.scene.image.Image;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import common.Properties;

/*
 * Die Klasse MainApp stellt die Hauptklasse der Clients dar. In ihr wird die GUI initialisiert 
 * und geladen sowie saemtliche "nach aussen" (zum Server) gerichteten Funktionalitaeten implementiert.
 */
public class MainApp extends Application {

    private Stage primaryStage;
    private BorderPane rootLayout;
    private RootLayoutController rootController;
    
    private AnchorPane enginePane, turntablePane, settingsPane, switchListPane;
    private EngineController engineController;
    private SwitchListController switchListController;
    private TurntableController turntableController;
    
    private Settings configuration;
    private SettingsController settingsController;
    
    private ObservableList<Tab> tabs = FXCollections.observableArrayList();
    private ObservableList<Switch> switches = FXCollections.observableArrayList();
    private ObservableList<Engine> engines = FXCollections.observableArrayList();
    private Switch turntable;
    
    private Socket client;
	private InputStream in;
	private OutputStream out;
	
	final ExecutorService UTILITY_THREAD_EXECUTOR;
	
	private Boolean connectionEstablished = false;
	private Boolean running = false;

	// Constructor
    public MainApp() {
    	UTILITY_THREAD_EXECUTOR = Executors.newFixedThreadPool(2);
    	
    	configuration = new Settings();    	
    	    	
        // Add the already known engines
        engines.add(new Engine("ICE", Properties.ICE_ID, "res/ICE 1.png"));
        engines.add(new Engine("Dampflok", Properties.LOK_ID, "res/BR 86.png"));
        engines.add(new Engine("Reichsbahn", Properties.REICHSBAHN_ID, "res/E 50.png"));
        
        // Add the already known switches
        for (int i=4;i<=15;i++){ 
        	if (i != 11 && i != 7) {
        		switches.add(new Switch("Weiche "+i, Properties.SWITCH4_ID+i-4));
        	}
        }
        
        turntable = new Switch("Drehteller", Properties.TURNTABLE_ID);
        
    }
    
    // Initializes the GUI
    @Override
    public void start(Stage primaryStage) {
        this.primaryStage = primaryStage;
        this.primaryStage.setTitle("Mrklin Control Client");
        this.primaryStage.getIcons().add(new Image("file:res/BO.png"));
        initRootLayout();
        initTabs();
    }

    // Loads and binds the root layouut
    public void initRootLayout() {
        try {
            // Load root layout from fxml file
            FXMLLoader loader = new FXMLLoader();
            loader.setLocation(MainApp.class.getResource("view/RootLayout.fxml"));
            rootLayout = (BorderPane) loader.load();
            rootController = loader.getController(); // only works if loader.load() has been called already
            rootController.setMainApp(this);
                      
            // Show the scene containing the root layout
            Scene scene = new Scene(rootLayout);
            primaryStage.setScene(scene);
            primaryStage.show();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    // Shows the control tabs inside the tab pane of the root layout.
    public void initTabs() {
        try {
        	FXMLLoader loader = new FXMLLoader();
            
        	loader.setLocation(MainApp.class.getResource("view/Engine.fxml"));
            enginePane = loader.load();
            engineController = loader.getController(); // only works if loader.load() has been called already
            engineController.setMainApp(this);
            
            FXMLLoader switchListLoader = new FXMLLoader();
            switchListLoader.setLocation(MainApp.class.getResource("view/SwitchList.fxml"));
            switchListPane = switchListLoader.load();
            switchListController = switchListLoader.getController();
            switchListController.setMainApp(this);
            switchListController.setSwitchControls();
            
            FXMLLoader turntableLoader = new FXMLLoader();
            turntableLoader.setLocation(MainApp.class.getResource("view/Turntable.fxml"));
            turntablePane = turntableLoader.load();
            turntableController = turntableLoader.getController();
            turntableController.setMainApp(this);
            
            FXMLLoader settingsLoader = new FXMLLoader();
            settingsLoader.setLocation(MainApp.class.getResource("view/Settings.fxml"));
            settingsPane = settingsLoader.load();
            settingsController = settingsLoader.getController();
            settingsController.setMainApp(this);
            settingsController.setSettings(configuration);
        	
        	// Add the tabs
        	tabs.add(new Tab("Loks", enginePane));
        	tabs.add(new Tab("Weichen", switchListPane));
        	tabs.add(new Tab("Drehteller", turntablePane));
        	tabs.add(new Tab("Einstellungen", settingsPane));
        	rootController.loadTabs();        	
           
        } catch (IOException e) {
            e.printStackTrace();
        }
        
    }
    
    public static void main(String[] args) {
        launch(args);
    }
  
    // Calls the root-controllers function setStatus to set the statusbar to the given string
    public void setStatus(String status){
    	rootController.setStatus(status);
    }
    
    // Returns the primary stage
    public Stage getPrimaryStage() {
        return primaryStage;
    }
    
    // Returns the tabs as an observable list
    public ObservableList<Tab> getTabs() {
        return tabs;
    }
    
    // Returns the engines as an observable list
    public ObservableList<Engine> getEngines() {
        return engines;
    }
    
    // Returns the switches as an observable list
    public ObservableList<Switch> getSwitches() {
        return switches;
    }
    
    // Returns the value of connectionEstablished
 	public Boolean getConnectionEstablished() {
 		return connectionEstablished;
 	}

    // Called, when the emegency-stop-button is pressed; calls emergencyStop and sets every engines speed to 0
	public void emergencyStopHandler() {
		emergencyStop();				
		for(Engine eng:engines){
			eng.setSpeed(0);
		}
		engineController.updateEngineStatus();
		setStatus("NOTAUS");
		
	}
	
	// Signalizes the server to cut the train-controls power or to re-enable it depending on its current state (running == false means the emergency-stop is active)
	public void emergencyStop () {
		if (connectionEstablished) {
			try {
				if (running) {
					byte[] datagram = {Properties.SEPERATOR, (byte) 0x00, (byte) 0x00, (byte) Properties.SYSTEM_STOP, (byte) 0x00, (byte) 0x00};
					out.write(datagram);
					running = false;
				}
				else {
					byte[] datagram = {Properties.SEPERATOR, (byte) 0x00, (byte) 0x00, (byte) Properties.SYSTEM_GO, (byte) 0x00, (byte) 0x00};
					out.write(datagram);
					running = true;
				}
			} catch (Exception e) {
				setStatus("Error at: emergencyStop");
				e.printStackTrace();
			}
		}
	}
	
	// Signalizes the server to set the given engines direction to the client-models direction-value
	public void setEngineDirection (Engine eng) {
		if (connectionEstablished) {
			try {
				byte dir; 
				if (eng.getDirection().get())
					dir = (byte) 0x01;
				else
					dir = (byte) 0x02;
				
				byte[] datagram = {Properties.SEPERATOR, (byte) ((eng.getMaerklinID().get()/(1<<8))%(1<<8)), (byte) (eng.getMaerklinID().get()%(1<<8)), (byte) Properties.ENGINE_SET_DIRECTION, dir, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: engineSetDirectionRight");
				e.printStackTrace();
			}
		}
	}
	
	// Signalizes the server to set the given engines direction to the client-models speed-value
	public void setEngineSpeed (Engine eng) {
		if (connectionEstablished) {
			try {
				int newEngineSpeed = eng.getSpeed().get();
				byte[] datagram = {Properties.SEPERATOR, (byte) ((eng.getMaerklinID().get()/(1<<8))%(1<<8)), (byte) (eng.getMaerklinID().get()%(1<<8)), (byte) Properties.ENGINE_SET_SPEED, (byte) ((newEngineSpeed/(1<<8))%(1<<8)), (byte) (newEngineSpeed%(1<<8))};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("engineSetSpeed");
				e.printStackTrace();
			}
		}
	}
	
	// Signalizes the server to set the given switch state to the client-models state-value
	public void setSwitchState (Switch sw) {
		if (connectionEstablished) {
			try {
				byte dir; 
				if (!sw.getState().get()) 	// Straight
					dir = (byte) 0x01;
				else						// Curved
					dir = (byte) 0x00;
				
				byte[] datagram = {Properties.SEPERATOR, (byte) ((sw.getMaerklinID().get()/(1<<8))%(1<<8)), (byte) (sw.getMaerklinID().get()%(1<<8)), (byte) Properties.SWITCH_SET_DIRECTION, dir, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: setSwitchState");
				e.printStackTrace();
			}
		}
	}
	
	
	// Signalizes the server to move the turntable one step in clockwise direction
	public void turntableTurnCW () {
		if (connectionEstablished) {
			try {
				byte[] datagram = {Properties.SEPERATOR, (byte) ((turntable.getMaerklinID().get()/(1<<8))%(1<<8)), (byte) (turntable.getMaerklinID().get()%(1<<8)), (byte) Properties.SWITCH_SET_DIRECTION, (byte) 0x00, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: turntableTurnCW");
				e.printStackTrace();
			}
		}
	}
	
	// Signalizes the server to move the turntable one step in counterclockwise direction
	public void turntableTurnCCW () {
		if (connectionEstablished) {
			try {
				byte[] datagram = {Properties.SEPERATOR, (byte) ((turntable.getMaerklinID().get()/(1<<8))%(1<<8)), (byte) (turntable.getMaerklinID().get()%(1<<8)), (byte) Properties.SWITCH_SET_DIRECTION, (byte) 0x01, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: turntableTurnCCW");
				e.printStackTrace();
			}
		}
	}	
	
	// Submits a thread to the UTILITY_THREAD_EXECUTOR to establish the connection to the server or closes the active connection depending on the value of connectionEstablished
	public void establishConnetion () {
		if (connectionEstablished) {
			try {
				byte[] datagram = {Properties.SEPERATOR, (byte) 0x00, (byte) 0x00, (byte) Properties.SESSION_ABORT, (byte) 0x00, (byte) 0x00};
				out.write(datagram);
				client.close();
				connectionEstablished = false;
				running = false;
				settingsController.updateSettingsStatus();
				
			} catch (Exception e) {
				setStatus("Error at: establishConnection (shutting down connection)");
				e.printStackTrace();
			}
		}
		else {
			UTILITY_THREAD_EXECUTOR.submit((new EstablishConnection(this)));
		}
	}
	
	/*
	 *  Diese Klasse stellt die Verbindung zum Server her und hält die Input- und Output-Streams 
	 *  offen. Weiterhin updated sie bei bestehender Verbindung zyklisch die GUI-Elemente mit den 
	 *  Werten der (clientseitig) zugehörigen Modelle.
	 */
	class EstablishConnection implements Runnable {
		
		// Serves to solve relations to the calling instance
		private final MainApp CONTROLLER_INSTANCE;
		
		EstablishConnection (MainApp controller) {
			CONTROLLER_INSTANCE = controller;
		}
		
		// Updates all GUI-elements to match the models state
		public void updateGUI () {
			engineController.updateEngineStatus();
			for (Switch sw : switches){
				sw.getController().get().updateSwitchStatus();
			}
		}
		
		// Establishes the connection to the server if no connection is established yet, submits a update-thread to the UTILITY_THREAD_EXECUTOR (meaning that updates from the server are processed from now on) and cyclically updates all GUI-elements through updateGUI
		@Override
		public void run() {
			if (!connectionEstablished) {
				try (Socket socket = new Socket(configuration.getSocketAddress().get().getHostName(), configuration.getSocketAddress().get().getPort())) {
					client = socket;
					client.setKeepAlive(true);
					client.setReuseAddress(true);
					in = client.getInputStream();
					out = client.getOutputStream();
					connectionEstablished = true;
					running = true;
					UTILITY_THREAD_EXECUTOR.submit((new UpdateFunctionality(CONTROLLER_INSTANCE, in)));
					Platform.runLater(new Runnable() {

						// Sets the connection-esttablishing-buttons text to "Verbindung trennen"
						@Override
						public void run() {
							CONTROLLER_INSTANCE.setStatus("Verbindung zu "+client.getRemoteSocketAddress()+" aufgebaut!");
							CONTROLLER_INSTANCE.settingsController.updateSettingsStatus();
						}
						
					});
					while (connectionEstablished) {
						Platform.runLater(new Runnable() {

							@Override
							public void run() {
								updateGUI();					
							}
						
						});
						Thread.sleep(50);	// Frees the resources between the GUI-updates
					}
				}
				catch (Exception e) {
					connectionEstablished = false;	// Interrupts the connection to the server in case of an error
					running = false;
					Platform.runLater(new Runnable() {

						// Sets the connection-esttablishing-buttons text to "Verbindung herstellen"
						@Override
						public void run() {
							CONTROLLER_INSTANCE.setStatus("Error at: establishConnection (establishing connection)");
							CONTROLLER_INSTANCE.settingsController.updateSettingsStatus();
						}
						
					});
					e.printStackTrace();
				}
				finally {
					if (connectionEstablished) {
						try {
							// Properly shuts down the connection to the server in case of an unexpected termination of the client-instance
							byte[] datagram = {Properties.SEPERATOR, (byte) 0x00, (byte) 0x00, (byte) Properties.SESSION_ABORT, (byte) 0x00, (byte) 0x00};
							out.write(datagram);
							client.close();
							connectionEstablished = false;
							running = false;
							Platform.runLater(new Runnable() {
	
								// Sets the connection-esttablishing-buttons text to "Verbindung herstellen"
								@Override
								public void run() {
									CONTROLLER_INSTANCE.settingsController.updateSettingsStatus();
								}
								
							});
						} catch (Exception e) {
							setStatus("Error at: establishConnection (shutting down connection)");
							e.printStackTrace();
						}
					}		
				}
			}
		}
	}

	/*
	 * Diese Runnable-Klasse lauscht auf dem Update-Port des Servers auf Status-Updates, 
	 * decodiert diese bei Empfang und updatet die (clientseitigen) Modelle auf die erhaltenen 
	 * Werte. Achtung: die Modelle werden geupdatet, nicht die GUI, da sich ansonsten Nebenläufigkeits-
	 * Problematiken ergeben können.
	 */
	class UpdateFunctionality implements Runnable {
		
		// Serves to solve relations to the calling instance
		private final MainApp CONTROLLER_INSTANCE;
		
		// Due to its length, the incoming datagram can't be handled in one step. Therefore a BufferedInputStream is chosen to buffer it until everything is properly processed.
		private final BufferedInputStream INPUT_STREAM;
		
		UpdateFunctionality (MainApp controller, InputStream in) {
			CONTROLLER_INSTANCE = controller;
			INPUT_STREAM = new BufferedInputStream(in, Properties.OUT_BUFFER_SIZE);
		}
		
		// Looks up the correct model related to the datagram and updates it to the received values
		public void parseDatagram (byte[] datagram) {
			if (datagram.length < 5) {	// Checks the given datagrams format
				setStatus("Error while updating!");
				return;
			}
			for (Engine eng : engines){	// Checks all engines
				if(eng.getMaerklinID().get()==((datagram[0]&0xFF)*(1<<8)+(datagram[1]&0xFF))){
					int engSpeed = (datagram[2]&0xFF)*(1<<8)+(datagram[3]&0xFF);
					boolean engDirection = (datagram[4] == (0x01));
					if (engSpeed != eng.getSpeed().get()) {
						eng.setSpeed(engSpeed);
					}
					if (engDirection != eng.getDirection().get()) {
						eng.setDirection(engDirection);
					}
					if (eng.equals(engineController.getSelectedEngine())){}				
				}
			}
			for (Switch sw : switches){	// Checks all switches
				if(sw.getMaerklinID().get()==((datagram[0]&0xFF)*(1<<8)+(datagram[1]&0xFF))){
					boolean swState = !(datagram[4]  == (byte) 0x01);
					if (swState != sw.getState().get()) {
						sw.setState(swState);
					}
				}
			}
		}
		
		
		// If a datagram is available, it will be scanned for its start (marked through two seperator-bytes right after another). Afterwards the data is read until the next seperator-byte or the datagram ends, the datas format is checked and, if everything is okay, the (clients) models are updated through parseDatagram.
		@Override
		public void run() {
			int stopCounter, dataCounter, counter;
			byte buffer[] = new byte[Properties.OUT_BUFFER_SIZE];
			byte datagram[] = new byte[Properties.IN_BUFFER_SIZE];
			boolean startByte;
			try {
				while (connectionEstablished) {
					stopCounter = 0;
					dataCounter = 0;
					counter = 0;
					startByte = false;
					while (INPUT_STREAM.available() > 0 && stopCounter < 5) {	// Checks if a datagram is available and not processed yet
						buffer[counter%Properties.OUT_BUFFER_SIZE] = (byte) INPUT_STREAM.read();
						if (stopCounter == 2) {	// Checks for the datagrams start
							startByte = true;
						}
						if (buffer[counter%Properties.OUT_BUFFER_SIZE] == Properties.SEPERATOR && dataCounter != 4) {	// Handles seperator-bytes
							dataCounter = 0;
							stopCounter++;
						}
						else {	// Handles data-bytes
							stopCounter = 0;
							dataCounter++;
							if (dataCounter == 5 && startByte) {	// If a full datagram consisting of five bytes is read, parseDatagram is called to handle the data
								dataCounter = 0;
								for (int i = 0; i < Properties.IN_BUFFER_SIZE; i++) {
									datagram[i] = buffer[counter-Properties.IN_BUFFER_SIZE+i+1];
								}
								parseDatagram(datagram);
							}
							else if (dataCounter > 5 && startByte) {	// Checks if the datagrams have the correct format
								throw (new Exception("Wrong data-update-format!"));
							}
						}
						counter++;
					}
					while (INPUT_STREAM.available() != 0) {	// Flushes the remaining buffer of the BufferedInputStream (doesn't contain any information), so that the next datagram will start at position 0 again (speeds up the next update-process) 
						INPUT_STREAM.read();
					}
					Thread.sleep(50);	// Frees the resources between the status-updates 
				}
			}
			catch (Exception e) {
				e.printStackTrace();
				Platform.runLater(new Runnable() {

					@Override
					public void run() {
						CONTROLLER_INSTANCE.setStatus("Error while updating!");						
					}
					
				});
			}
		}
	}
}