package server;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

//import javax.xml.bind.DatatypeConverter;

import common.Properties;

/*
 * MaerklinServer stellt die Hauptklasse des Servers dar. Alle verwendeten 
 * Objekte und Services (bspw. Executer-Services) werden initialisiert und
 * der Server-Thread wird gestartet. Weiterhin implementiert diese Klasse 
 * das Maerklin-Protokoll und dient damit fuer alle anderen abgeleiteten oder
 * erzeugten Objekte als Referenz fuer die Kommunikation mit dem Server. 
 * Grundlegende Funktionen wie der Notaus werden direkt von dieser Klasse 
 * umgesetzt.
 */
public class MaerklinServer{
	
	// Declaration of objects
	
	// List to monitor the registered clients
	ArrayList<Socket> clients;
		
	// Control-elements and thread to handle new clients
	private MaerklinProtocol protocol;
	private UDPListener overwatch;
	private ServerThread serverThread;
	
	// Executor-services to handle the threads
	final ExecutorService SERVER_THREAD_EXECUTOR; 
	final ExecutorService CLIENT_THREAD_EXECUTOR;
	final ExecutorService UTILITY_THREAD_EXECUTOR;
	
	// Accessories
	private Engine ice;
	private Engine lok;
	private Engine reichsbahn;
		
	private Switch switch4;
	private Switch switch5;
	private Switch switch6;
	private Switch switch8;
	private Switch switch9;
	private Switch switch10;
	private Switch switch12;
	private Switch switch13;
	private Switch switch14;
	private Switch switch15;
	private Switch turntable;
	
	// Constructor
	public MaerklinServer(){
		
		// Initializing
		System.out.print("Maerklin Control Server\nLukas Friedrichsen, Philipp Stenkamp; (c) 2017\n\nInitializing...\n");
		
		clients = new ArrayList<Socket>();
				
		protocol = new MaerklinProtocol();
		overwatch = new UDPListener();
		serverThread = new ServerThread(this);
		
		SERVER_THREAD_EXECUTOR = Executors.newFixedThreadPool(2);	// Server- and listener-thread
		CLIENT_THREAD_EXECUTOR = Executors.newCachedThreadPool();	// Client-threads
		UTILITY_THREAD_EXECUTOR = Executors.newCachedThreadPool();	// Handle- and update-threads
		
		ice = new Engine();
		lok = new Engine();
		reichsbahn = new Engine();
		
		switch4 = new Switch();
		switch5 = new Switch();
		switch6 = new Switch();
		switch8 = new Switch();
		switch9 = new Switch();
		switch10 = new Switch();
		switch12 = new Switch();
		switch13 = new Switch();
		switch14 = new Switch();
		switch15 = new Switch();
		turntable = new Switch();
		
		// Setting IDs
		ice.setEngineID(Properties.ICE_ID);
		lok.setEngineID(Properties.LOK_ID);
		reichsbahn.setEngineID(Properties.REICHSBAHN_ID);
		
		switch4.setSwitchID(Properties.SWITCH4_ID);
		switch5.setSwitchID(Properties.SWITCH5_ID);
		switch6.setSwitchID(Properties.SWITCH6_ID);
		switch8.setSwitchID(Properties.SWITCH8_ID);
		switch9.setSwitchID(Properties.SWITCH9_ID);
		switch10.setSwitchID(Properties.SWITCH10_ID);
		switch12.setSwitchID(Properties.SWITCH12_ID);
		switch13.setSwitchID(Properties.SWITCH13_ID);
		switch14.setSwitchID(Properties.SWITCH14_ID);
		switch15.setSwitchID(Properties.SWITCH15_ID);
		turntable.setSwitchID(Properties.TURNTABLE_ID);
		
		// Starting the system and setting ("safe") values
		try {
			start();
			initialize();
		}
		catch (Exception e){
			System.out.println(e.getMessage());
			System.exit(1);
		}
	}
	
	// Emergency stop; sets the speed of all engines to 0 out of safety reasons
	public void emergencyStop() throws Exception {
		protocol.systemStop();
		ice.setEngineSpeed(0);
		lok.setEngineSpeed(0);
		reichsbahn.setEngineSpeed(0);
	}
	
	// Resets emergency stop
	public void start() throws Exception {
		protocol.systemGo();
	}
	
	// Initialize; setting ("safe") values (engine speed to 0 and switch direction to straight)
	public void initialize() throws Exception {
		ice.setEngineDirection(1);
		ice.setEngineSpeed(0);
		
		lok.setEngineDirection(1);
		lok.setEngineSpeed(0);
		
		reichsbahn.setEngineDirection(1);
		reichsbahn.setEngineSpeed(0);
		
		switch4.setSwitchDirection(1);
		switch5.setSwitchDirection(1);
		switch6.setSwitchDirection(1);
		switch8.setSwitchDirection(1);
		switch9.setSwitchDirection(1);
		switch10.setSwitchDirection(1);
		switch12.setSwitchDirection(1);
		switch13.setSwitchDirection(1);
		switch14.setSwitchDirection(1);
		switch15.setSwitchDirection(1);
	}
		
	// Starts the listener-thread
	public void startListener(){
		SERVER_THREAD_EXECUTOR.submit(overwatch);
	}
	
	// Listens to the client-port and initializes a new Thread to handle the client-messages
	public void listen(){
		SERVER_THREAD_EXECUTOR.submit(serverThread);
	}
	
	// Starts a thread to broadcast the current status for every registered client
	public void statusUpdate(){
		for (Socket clientSocket : clients) {
			try {
				UTILITY_THREAD_EXECUTOR.submit(new UpdateThread(clientSocket.getOutputStream()));
			} 
			catch (IOException e) {
				System.out.println("Failed to receive output-stream of one of the clients!");
			}
		}
	}
	
}

/*
 * Diese Klasse oeffnet einen Server-Port und erstellt einen neue Client-Thread-Instanz
 * fuer jeden Client, der sich neu an dem Server anmeldet.
 */
class ServerThread implements Runnable {
	
	// Declaration of objects
	
	// Instance of the server; serves to solve references
	private final MaerklinServer SERVER_INSTANCE;
	
	// Constructor
	ServerThread (MaerklinServer server) {
		SERVER_INSTANCE = server;
	}

	// Waits for clients to connect and submits a new client-thread afterwards
	@Override
	public void run() {
		try(ServerSocket serverSocket = new ServerSocket(Properties.PORT)){
			serverSocket.setReuseAddress(true);
			System.out.println("Server-socket established!");
			while (true) {
				Socket socket = serverSocket.accept();
				System.out.println("Connection to client "+socket.getRemoteSocketAddress()+" established!");
				SERVER_INSTANCE.CLIENT_THREAD_EXECUTOR.submit(new ClientThread(SERVER_INSTANCE, socket, socket.getInputStream()));
			}
		} 
		catch (Exception e) {
			System.out.println("Accessing user-interaction-port failed! Trying again!");
			run();
		}
		finally {
			// Properly shuts down all running threads if the server is closed
			SERVER_INSTANCE.SERVER_THREAD_EXECUTOR.shutdownNow();
			SERVER_INSTANCE.CLIENT_THREAD_EXECUTOR.shutdownNow();
			SERVER_INSTANCE.UTILITY_THREAD_EXECUTOR.shutdownNow();
		}
	}
	
}

/*
 *  Die Klasse ClientThread verwaltet die Kommunikation mit dem Client
 *  im Hintergrund und oeffnet bei eingehenden Befehlen einen neuen Handling-Thread.
 */
class ClientThread implements Runnable {
	
	// Declaration of objects
	
	// Instance of the server; serves to solve references
	private final MaerklinServer SERVER_INSTANCE;
	
	// Instance of the client-socket; serves to solve references
	private final Socket CLIENT;
	
	// Wrapper for the client-sockets' input-stream to buffer the incoming data
	private final BufferedInputStream IN_STREAM;
	
	byte[] data;
	
	// Constructor
	ClientThread (MaerklinServer server, Socket socket, InputStream in) {
		CLIENT = socket;
		IN_STREAM = new BufferedInputStream(in);
		SERVER_INSTANCE = server;
		SERVER_INSTANCE.clients.add(socket);
		data = new byte[Properties.IN_BUFFER_SIZE];
	}
	
	// Checks if the client sends a session-abort-message passes client-requests to a handle-thread
	@Override
	public void run() {
		int buffer;
		while (true) {
			try {
				if (IN_STREAM.available() != 0) {
					buffer = IN_STREAM.read();	// Reads one byte from the input-stream
					if ((byte) buffer == (byte) Properties.SESSION_ABORT) {	// Checks for a session-abort signal and potentially severs the connection potentially
						SERVER_INSTANCE.clients.remove(CLIENT);
						CLIENT.close();
						System.out.println("Conenction to client "+CLIENT.getRemoteSocketAddress()+" aborted!");
						return;
					}
					else if ((byte) buffer == (byte) Properties.SEPERATOR) {	// Checks for the beginning of a new datagram (all datagrams begin with a seperator-byte to mark their start)
						for (int i = 0; i < Properties.IN_BUFFER_SIZE; i++) {	// Tries to read the whole datagram
							buffer = IN_STREAM.read();
							if ((byte) buffer == (byte) Properties.SESSION_ABORT) {	// Again checks for a session-abort signal from the client
								SERVER_INSTANCE.clients.remove(CLIENT);
								CLIENT.close();
								System.out.println("Connection to client "+CLIENT.getRemoteSocketAddress()+" aborted!");
								return;
							}
							else {
								data[i] = (byte) (buffer%(1<<8));	// Typecast to byte through bitshift
							}
						}
						SERVER_INSTANCE.UTILITY_THREAD_EXECUTOR.submit(new HandleThread(SERVER_INSTANCE, data));	// Submits a new handle-thread for the datagram
					}
				}
				else {
					Thread.sleep(10);	// Frees the resources until the next check
				}
			}
			catch (Exception e) {
				// Unregister the client if an error occurers
				System.out.println("An error occured while reading the clients input-stream!");
				SERVER_INSTANCE.clients.remove(CLIENT);
				System.out.println("Conenction to client "+CLIENT.getRemoteSocketAddress()+" aborted!");
				return;
			}
		}
	}
	
}

/*
 * HandleThread verarbeitet Anfragen von den Clients und initiiert entsprechende Aktionen.
 * Format des Datagrams: 	1.+2. Byte: Maerklin-ID
 * 							3. Byte: Befehl
 * 							4.+5. Byte: potentielle Parameter
 */
class HandleThread implements Runnable {
	
	// Declaration of objects// Declaration of objects
	
	// Instance of the server; serves to solve references
	private final MaerklinServer SERVER_INSTANCE;
	
	byte[] data;
	
	// Constructor
	HandleThread (MaerklinServer server, byte[] incomingData) {
		SERVER_INSTANCE = server;
		data = incomingData;
	}

	// Reads the command and initiates the according action
	@Override
	public void run() {
		try {
			switch (data[2]) {	// Compare the command-byte to the values set in Properties
				case Properties.SYSTEM_STOP:	// Emergency-stop (cuts the power)
					//System.out.println("System stop");
					SERVER_INSTANCE.emergencyStop();
					break;
				case Properties.SYSTEM_GO:	// Resets the emergency-stop (re-enables power)
					//System.out.println("System go");
					SERVER_INSTANCE.start();
					break;
				case Properties.GET_STATUS:	// Status-update
					//System.out.println("Status update");
					SERVER_INSTANCE.statusUpdate();
					break;
				case Properties.ENGINE_SET_SPEED:	// Set engine-speed to the given value
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == ((data[0]&0xFF)*(1<<8)+(data[1]&0xFF))) {
							//System.out.println("Set engine speed");
							train.setEngineSpeed((data[3]&0xFF)*(1<<8)+(data[4]&0xFF));
							break;
						}
					}
					break;
				case Properties.ENGINE_INCREASE_SPEED:	// Increase the engine-speed by 50
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == ((data[0]&0xFF)*(1<<8)+(data[1]&0xFF))) {
							//System.out.println("Increase engine speed");
							train.increaseEngineSpeed();
							break;
						}
					}
					break;
				case Properties.ENGINE_DECREASE_SPEED:	// Decrease the engine-speed by 50
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == ((data[0]&0xFF)*(1<<8)+(data[1]&0xFF))) {
							//System.out.println("Decrease engine speed");
							train.decreaseEngineSpeed();
							break;
						}
					}
					break;
				case Properties.ENGINE_SET_DIRECTION:	// Set engine-direction to the given value
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == ((data[0]&0xFF)*(1<<8)+(data[1]&0xFF))) {
							//System.out.println("Set engine direction");
							train.setEngineDirection(data[3]);
							break;
						}
					}
					break;
				case Properties.ENGINE_SWITCH_DIRECTION:	// Switch the engine-direction
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == ((data[0]&0xFF)*(1<<8)+(data[1]&0xFF))) {
							//System.out.println("Switch engine direction");
							train.setEngineDirection(3);
							break;
						}
					}
					break;
				case Properties.SWITCH_SET_DIRECTION:	// Set the switch-direction to the given value or move the turntable clockwise or counterclockwise
					for (Switch sw : Switch.switches) {
						if (sw.getSwitchID() == ((data[0]&0xFF)*(1<<8)+(data[1]&0xFF))) {
							//System.out.println("Set switch direction");
							sw.setSwitchDirection(data[3]);
							break;
						}
					}
					break;
				case Properties.SWITCH_SWITCH_DIRECTION:	// Switch the switch-direction
					for (Switch sw : Switch.switches) {
						if (sw.getSwitchID() == ((data[0]&0xFF)*(1<<8)+(data[1]&0xFF))) {
							//System.out.println("Switch switch direction");
							sw.changeSwitchDirection();
							break;
						}
					}
					break;
				default:	// Command doesn't match any of the cases
					System.out.println("No viable command!\n");
			}
		}
		catch (Exception e) {
			System.out.println(e.getMessage());
		}
	}
	
}

/*
 * Diese Klasse konvertiert bei Aufruf den aktuellen Status der (serverseitigen) 
 * Modelle in Byte-Format und wandelt diese unter Hinzufuegen von Start- End- und 
 * Trenn-Bytes in ein Datagram um, dass an den Client gesendet wird.
 */
class UpdateThread implements Runnable {
	
	// Declaration of objects
	
	// Output-stream of the client handled by the thread
	private final OutputStream OUT_STREAM;
	
	// Buffer for the datagram
	private byte[] outgoingData;
	
	// Constructor
	UpdateThread (OutputStream out) {
		
		// Setting the broadcast-port
		OUT_STREAM = out;
	
		// Initializing the buffer-array to the size of the datagram
		outgoingData = new byte[Properties.OUT_BUFFER_SIZE];
	}
	
	// Reads every registered objects data into outgoingData and sends the status
	@Override
	public void run() {
		int position = 0;
		try {
			outgoingData[position++] = Properties.SEPERATOR;	// Start-bytes
			outgoingData[position++] = Properties.SEPERATOR;
			for (Engine train : Engine.engines) {
				if (position >= Properties.OUT_BUFFER_SIZE-9) {
					Exception e = new Exception("Overload of registered elements! Update-buffer too small!");
					throw e;
				}
				outgoingData[position++] = (byte) ((train.getEngineID()/(1<<8))%(1<<8));	// Maerklin-ID
				outgoingData[position++] = (byte) ((train.getEngineID()%(1<<8)));
				outgoingData[position++] = (byte) ((train.getEngineSpeed()/(1<<8))%(1<<8));	// Speed
				outgoingData[position++] = (byte) ((train.getEngineSpeed()%(1<<8)));
				outgoingData[position++] = (byte) ((train.getEngineDirection()%(1<<24)));	// Direction
				outgoingData[position++] = Properties.SEPERATOR;	// Separator-byte between the engines
			}
			outgoingData[position++] = Properties.SEPERATOR;	// Separator-bytes between the engines and the switches
			outgoingData[position++] = Properties.SEPERATOR;
			outgoingData[position++] = Properties.SEPERATOR;
			for (Switch sw : Switch.switches) {
				if (position >= Properties.OUT_BUFFER_SIZE-10) {
					Exception e = new Exception("Overload of registered elements! Update-buffer too small!");
					throw e;
				}
				outgoingData[position++] = (byte) ((sw.getSwitchID()/(1<<8))%(1<<8));	// Maerklin-ID
				outgoingData[position++] = (byte) ((sw.getSwitchID()%(1<<8)));
				outgoingData[position++] = (byte) 0x00;	// Inserted two null-bytes to match the engines-format
				outgoingData[position++] = (byte) 0x00;
				outgoingData[position++] = (byte) ((sw.getSwitchDirection()%(1<<24)));	// Direction
				outgoingData[position++] = Properties.SEPERATOR;	// Separator-byte between the switches
			}
			outgoingData[position++] = Properties.SEPERATOR;	// Stop-bytes
			outgoingData[position++] = Properties.SEPERATOR;
			outgoingData[position++] = Properties.SEPERATOR;
			outgoingData[position++] = Properties.SEPERATOR;
		}
		catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		//System.out.printf("\nSende: " + DatatypeConverter.printHexBinary(outgoingData));
		try {
			OUT_STREAM.write(outgoingData, 0, Properties.OUT_BUFFER_SIZE);	// Send the datagram to the client
		}
		catch (Exception e) {
			System.out.println("Communication-error whilst status-update!");
		}
	}
	
}
