package server;

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import javax.xml.bind.DatatypeConverter;

/*
 * Der UDPListener ist in der Lage auf den Ausgabe-Port der Maerklin-C2-Steuerung 
 * zu lauschen, ausgehende Datagramme zu verarbeiten und die serverseitigen Modelle
 * zu updaten. Es ist sowohl eine "manuelle" Version wie auch eine Automatik (per 
 * Runnable) implementiert.
 */
public class UDPListener implements Runnable{
	
	// Defining the size of the datagrams received from the Maerklin-control-station
	public final static int MAERKLIN_DATAGRAM_SIZE = 13;
	
	// Setting the output-port of the maerklin-control-station
	private final static int port = 15730;
	
	private byte[] in;
	
	// Constructor
	public UDPListener(){
		in = new byte[MAERKLIN_DATAGRAM_SIZE];
	}
	
	// Listens for answers from the control-station and writes it to the array in
	public void listen(){
		try(DatagramSocket socket = new DatagramSocket(port)){
			socket.setReuseAddress(true);
			DatagramPacket incomingData = new DatagramPacket(in, in.length);
			socket.receive(incomingData);
			in = incomingData.getData();
		} 
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	// Returns the last received datagram
	public byte[] getData(){
		return in;
	}
	
	// Decodes the engine-speed from the datagram
	public int getEngineSpeed(){
		return ((in[9]&0xFF)*(1<<8)+(in[10]&0xFF));
	}
	
	
	// Decodes the engine-direction from the datagram
	public int getEngineDirection(){
		return (in[9]&0xFF);
	}
	
	// Decodes the switch-direction from the datagram
	public int getSwitchDirection(){
		return (in[9]&0xFF);
	}
	
	// Decodes the Maerklin-ID from the datagram
	public int getAddress(){
		return ((in[7]&0xFF)*(1<<8)+(in[8]&0xFF));
	}
	
	// Decodes the command from the datagram
	public int getCommand(){
		return ((in[1]&0xFF)&0xFE);
	}
	
	// Checks if the last received datagram is an answer from the control-station
	public boolean isAnswer(){
		if ((in[1]&0x01) == 1){
			return true;
		}
		else {
			return false;
		}
	}
	
	// Prints the last received datagram in human-readable form
	public void printData(byte[] data){
		try{
			byte[] kennung = new byte[4];
			byte[] laenge = new byte[1];
			byte[] nutzdaten = new byte[8];
			System.arraycopy(data, 0, kennung, 0, 4);
			System.arraycopy(data, 4, laenge, 0, 1);
			System.arraycopy(data, 5, nutzdaten, 0, 8);
			System.out.print("\n\nKennung   DLC Nutzdaten\n" + DatatypeConverter.printHexBinary(kennung) + "  " + DatatypeConverter.printHexBinary(laenge) + "  " + DatatypeConverter.printHexBinary(nutzdaten));
		}
		catch(Exception e){
			e.printStackTrace();
		}
	}
	
	// Debugging-functionality
	public static void main(String args[]){
		UDPListener testListener = new UDPListener();
		while (true){
			testListener.listen();
			testListener.printData(testListener.getData());
		}
	}

	// Automatizes the listener-functionalit and updates the models' values with the received data
	@Override
	public void run() {
		//System.out.println("UDPListener start!");
		try(DatagramSocket socket = new DatagramSocket(port)){	// Tries to open a socket set to the output-port of the control-station
			socket.setReuseAddress(true);
			while (true){
				DatagramPacket incomingData = new DatagramPacket(in, in.length);
				socket.receive(incomingData);
				in = incomingData.getData();
				if (isAnswer()){	// Checks if the datagram is an answer
					if (getCommand() == 0x08){	// Engine-speed
						for (Engine train : Engine.engines){
							if (train.getEngineID() == getAddress()){
								train.engineSpeed = getEngineSpeed();
								// System.out.println("Speed of Engine "+train.getEngineID()+" changed to "+train.getEngineSpeed());
								break;
							}
						}
					}
					else if (getCommand() == 0x0A){	// Engine-direction
						for (Engine train : Engine.engines){
							if (train.getEngineID() == getAddress()){
								train.engineDirection = getEngineDirection();
								train.engineSpeed = 0;
								// System.out.println("Direction of Engine "+train.getEngineID()+" changed to "+train.getEngineDirection());
								break;
							}
						}
					}
					else if (getCommand() == 0x16){	// Switch-direction
						for (Switch sw : Switch.switches){
							if (sw.getSwitchID() == getAddress()){
								sw.switchDirection = getEngineDirection();
								// System.out.println("Direction of Switch "+sw.getSwitchID()+" changed to "+sw.getSwitchDirection());
								break;
							}
						}
					}
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
}
