package maerklinProject_ServerApplication;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;

import javax.xml.bind.DatatypeConverter;

import maerklinProject_Properties.ProjectProperties;

public class MaerklinServer{
	
	// Setting ports and defining size of the buffers
	private final static int IN_BUFFER_SIZE = 5;	// First and second byte: accessory-ID, third byte: command, fourth fifth byte: potential parameters
	private final static int OUT_BUFFER_SIZE = 1024;
	
	// Declaration of objects
	ArrayList<Socket> clients;
		
	private MaerklinProtocol protocol;
	private UDPListener overwatch;
	
	private Engine ice;
	private Engine lok;
	private Engine reichsbahn;
		
	private Switch switch4;
	private Switch switch5;
	private Switch switch6;
	private Switch switch8;
	private Switch switch9;
	private Switch switch10;
	private Switch switch12;
	private Switch switch13;
	private Switch switch14;
	private Switch switch15;
	private Switch turntable;
	
	public MaerklinServer(){
		
		// Initializing
		clients = new ArrayList<Socket>();
				
		protocol = new MaerklinProtocol();
		overwatch = new UDPListener();
		
		ice = new Engine();
		lok = new Engine();
		reichsbahn = new Engine();
		
		switch4 = new Switch();
		switch5 = new Switch();
		switch6 = new Switch();
		switch8 = new Switch();
		switch9 = new Switch();
		switch10 = new Switch();
		switch12 = new Switch();
		switch13 = new Switch();
		switch14 = new Switch();
		switch15 = new Switch();
		turntable = new Switch();
		
		// Setting IDs
		ice.setEngineID(ProjectProperties.ICE_ID);
		lok.setEngineID(ProjectProperties.LOK_ID);
		reichsbahn.setEngineID(ProjectProperties.REICHSBAHN_ID);
		
		switch4.setSwitchID(ProjectProperties.SWITCH4_ID);
		switch5.setSwitchID(ProjectProperties.SWITCH5_ID);
		switch6.setSwitchID(ProjectProperties.SWITCH6_ID);
		switch8.setSwitchID(ProjectProperties.SWITCH8_ID);
		switch9.setSwitchID(ProjectProperties.SWITCH9_ID);
		switch10.setSwitchID(ProjectProperties.SWITCH10_ID);
		switch12.setSwitchID(ProjectProperties.SWITCH12_ID);
		switch13.setSwitchID(ProjectProperties.SWITCH13_ID);
		switch14.setSwitchID(ProjectProperties.SWITCH14_ID);
		switch15.setSwitchID(ProjectProperties.SWITCH15_ID);
		turntable.setSwitchID(ProjectProperties.TURNTABLE_ID);
		
		// Starting the system and setting ("safe") values
		try {
			start();
			initialize();
		}
		catch (Exception e){
			System.out.println(e.getMessage());
			System.exit(1);
		}
	}
	
	// Emergency stop
	public void emergencyStop() throws Exception {
		protocol.systemStop();
		ice.setEngineSpeed(0);
		lok.setEngineSpeed(0);
		reichsbahn.setEngineSpeed(0);
	}
	
	// Emergency stop
	public void start() throws Exception {
		protocol.systemGo();
	}
	
	// Initialize; setting ("safe") values
	public void initialize() throws Exception {
		ice.setEngineDirection(0);
		ice.setEngineSpeed(0);
		
		lok.setEngineDirection(0);
		lok.setEngineSpeed(0);
		
		reichsbahn.setEngineDirection(0);
		reichsbahn.setEngineSpeed(0);
		
		switch4.setSwitchDirection(0);
		switch5.setSwitchDirection(0);
		switch6.setSwitchDirection(0);
		switch8.setSwitchDirection(0);
		switch9.setSwitchDirection(0);
		switch10.setSwitchDirection(0);
		switch12.setSwitchDirection(0);
		switch13.setSwitchDirection(0);
		switch14.setSwitchDirection(0);
		switch15.setSwitchDirection(0);
	}
		
	// Starts the listener-thread
	public void startListener(){
		overwatch.run();
	}
	
	// Listens to the client-port and initializes a new Thread to handle the client-messages
	public void listen(){
		try(ServerSocket serverSocket = new ServerSocket(ProjectProperties.PORT)){
			serverSocket.setReuseAddress(true);
			Socket socket = serverSocket.accept();
			(new ClientThread(this, socket, socket.getInputStream(), IN_BUFFER_SIZE)).run();
		} 
		catch (Exception e) {
			System.out.println("Accessing user-interaction-port failed! Trying again!");
		}
	}
	
	// Starts the thread to broadcast the current status
	public void statusUpdate(){
		for (Socket clientSocket : clients) {
			try {
				(new UpdateThread(clientSocket.getOutputStream(), OUT_BUFFER_SIZE)).run();
			} catch (IOException e) {
				System.out.println("Failed to receive output-stream of one of the clients!");
			}
		}
	}
	
}

// Handles the communication with the clients in the background
class ClientThread implements Runnable {
	
	// Declaration of objects
	private final Socket CLIENT;
	private final InputStream IN_STREAM;
	private final int IN_BUFFER_SIZE;
	private final MaerklinServer SERVER_INSTANCE;
	byte[] data;
	int[] dataRaw;
	int counter;
	
	// Constructor
	ClientThread (MaerklinServer server, Socket socket, InputStream in, int bufferSize) {
		CLIENT = socket;
		IN_STREAM = in;
		IN_BUFFER_SIZE = bufferSize;
		SERVER_INSTANCE = server;
		SERVER_INSTANCE.clients.add(socket);
		data = new byte[IN_BUFFER_SIZE];
		dataRaw = new int[IN_BUFFER_SIZE];
		counter = 0;
	}
	
	@Override
	public void run() {
		while (true) {
			try {
				dataRaw[counter] = IN_STREAM.read();
				if (dataRaw[counter++] == ProjectProperties.SESSION_ABORT) {
					SERVER_INSTANCE.clients.remove(CLIENT);
					CLIENT.close();
					return;
				}
				else if (counter == IN_BUFFER_SIZE) {
					for (int i = 0; i < IN_BUFFER_SIZE; i++){
						data[i] = (byte) (dataRaw[i]%(1<<8));
					}
					(new HandleThread(SERVER_INSTANCE, data)).run();
					counter = -1;
				}
			}
			catch (Exception e) {
				System.out.println("An error occured while reading client-input-stream!");
			}
		}
	}
	
}

// Processes the incoming datagram and initializes appropriate actions
class HandleThread implements Runnable {
	
	// Declaration of objects
	MaerklinServer serverInstance;
	byte[] data;
	
	// Constructor
	HandleThread (MaerklinServer server, byte[] incomingData) {
		serverInstance = server;
		data = incomingData;
	}

	@Override
	public void run() {
		try {
			switch (data[1]) {
				case ProjectProperties.SYSTEM_STOP:
					serverInstance.emergencyStop();
					break;
				case ProjectProperties.SYSTEM_GO:
					serverInstance.start();
					break;
				case ProjectProperties.GET_STATUS:
					serverInstance.statusUpdate();
					break;
				case ProjectProperties.ENGINE_SET_SPEED:
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == data[0]) {
							train.setEngineSpeed(data[3]<<8+data[4]);
							break;
						}
					}
					break;
				case ProjectProperties.ENGINE_INCREASE_SPEED:
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == data[0]) {
							train.increaseEngineSpeed();
							break;
						}
					}
					break;
				case ProjectProperties.ENGINE_DECREASE_SPEED:
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == data[0]) {
							train.decreaseEngineSpeed();
							break;
						}
					}
					break;
				case ProjectProperties.ENGINE_SET_DIRECTION:
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == data[0]) {
							train.setEngineDirection(data[3]);
							break;
						}
					}
					break;
				case ProjectProperties.ENGINE_SWITCH_DIRECTION:
					for (Engine train : Engine.engines) {
						if (train.getEngineID() == data[0]) {
							train.setEngineDirection(3);
							break;
						}
					}
					break;
				case ProjectProperties.SWITCH_SET_DIRECTION:
					for (Switch sw : Switch.switches) {
						if (sw.getSwitchID() == data [0]) {
							sw.setSwitchDirection(data[3]);
							break;
						}
						break;
					}
				case ProjectProperties.SWITCH_SWITCH_DIRECTION:
					for (Switch sw : Switch.switches) {
						if (sw.getSwitchID() == data [0]) {
							sw.changeSwitchDirection();
							break;
						}
						break;
					}
				default:
					System.out.println("No viable command!");
			}
		}
		catch (Exception e) {
			System.out.println(e.getMessage());
		}
	}
	
}

//Broadcasts the current status
class UpdateThread implements Runnable {
	
	// Declaration of objects
	private final OutputStream OUT_STREAM;
	private final int OUT_BUFFER_SIZE;
	private byte[] outgoingData;
	
	// Constructor
	UpdateThread (OutputStream out, int bufferSize) {
		
		// Setting the broadcast-port and the outgoing-data-buffer-size
		OUT_STREAM = out;
		OUT_BUFFER_SIZE = bufferSize;
	
		// Initializing
		outgoingData = new byte[OUT_BUFFER_SIZE];
	}
	
	// Reads every registered objects data into outgoingData and broadcasts the status
	@Override
	public void run() {
		int position = 0;
		try {
			outgoingData[position++] = (byte) 0xff;
			outgoingData[position++] = (byte) 0xff;
			for (Engine train : Engine.engines) {
				if (position <= 1016) {
					Exception e = new Exception("Overload of registered elements! Update-buffer too small!");
					throw e;
				}
				outgoingData[position++] = Byte.parseByte(""+Math.floor((train.getEngineID()/(1<<8))%(1<<8)));
				outgoingData[position++] = Byte.parseByte(""+Math.floor(train.getEngineID()%(1<<8)));
				outgoingData[position++] = Byte.parseByte(""+Math.floor((train.getEngineSpeed()/(1<<8))%(1<<8)));
				outgoingData[position++] = Byte.parseByte(""+Math.floor(train.getEngineSpeed()%(1<<8)));
				outgoingData[position++] = Byte.parseByte(""+Math.floor(train.getEngineDirection()%(1<<24)));
				outgoingData[position++] = (byte) 0xff;
			}
			outgoingData[position++] = (byte) 0xff;
			outgoingData[position++] = (byte) 0xff;
			outgoingData[position++] = (byte) 0xff;
			for (Switch sw : Switch.switches) {
				if (position >= 1018) {
					Exception e = new Exception("Overload of registered elements! Update-buffer too small!");
					throw e;
				}
				outgoingData[position++] = Byte.parseByte(""+Math.floor((sw.getSwitchID()/(1<<8))%(1<<8)));
				outgoingData[position++] = Byte.parseByte(""+Math.floor(sw.getSwitchID()%(1<<8)));
				outgoingData[position++] = Byte.parseByte(""+Math.floor(sw.getSwitchDirection()%(1<<24)));
				outgoingData[position++] = (byte) 0xff;
			}
			outgoingData[position++] = (byte) 0xff;
			outgoingData[position++] = (byte) 0xff;
			outgoingData[position++] = (byte) 0xff;
			outgoingData[position++] = (byte) 0xff;
		}
		catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
		}
		System.out.printf("\nSende: " + DatatypeConverter.printHexBinary(outgoingData));
		try {
			OUT_STREAM.write(outgoingData, 0, OUT_BUFFER_SIZE);
		}
		catch (Exception e) {
			System.out.println("Communication-error whilst status-update!");
		}
	}
	
}
