package gui;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;

import gui.MainApp.updateThread;
import gui.model.Engine;
import gui.model.Switch;
import gui.model.Settings;
import gui.view.EngineController;
//import gui.view.TurntableController;
import gui.view.RootLayoutController;
import gui.view.SettingsController;
import gui.view.SwitchListController;
import javafx.application.Application;
import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.Tab;
import javafx.scene.image.Image;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import common.Properties;

public class MainApp extends Application {

    private Stage primaryStage;
    private BorderPane rootLayout;
    private RootLayoutController rootController;
    
    private AnchorPane enginePane, turntablePane, settingsPane, switchListPane;
    private EngineController engineController;
    private SwitchListController switchListController;
//    private TurntableController turntableController;
    
    private Settings configuration;
    private SettingsController settingsController;
    
    private ObservableList<Tab> tabs = FXCollections.observableArrayList();
    private ObservableList<Switch> switches = FXCollections.observableArrayList();
    private ObservableList<Engine> engines = FXCollections.observableArrayList();
    private Switch turntable;
    
    private Socket client;
	private InputStream in;
	private OutputStream out;
	
	private Boolean connectionEstablished = false;
      
    /**
     * Constructor
     */
    public MainApp() {
    	configuration = new Settings();    	
    	    	
        // Add the already known engines
        engines.add(new Engine("ICE", 0x0002, "res/ICE 1.png"));
        engines.add(new Engine("Dampflok", 0x4005, "res/BR 86.png"));
        engines.add(new Engine("Reichsbahn", 0x4007, "res/E 50.png"));
        
        // Add the already known switches // yes, lazy, i know
        for (int i=4;i<=15;i++){ switches.add(new Switch("Weiche "+i, 0x3000+(i-1))); }
        
        turntable = new Switch("Drehteller", Properties.TURNTABLE_ID);
        
    }

    /**
     * Returns the objects as an observable list. 
     * @return
     */
    public ObservableList<Tab> getTabs() {
        return tabs;
    }
    
    public ObservableList<Engine> getEngines() {
        return engines;
    }
    
    public ObservableList<Switch> getSwitches() {
        return switches;
    }
    
    
    @Override
    public void start(Stage primaryStage) {
        this.primaryStage = primaryStage;
        this.primaryStage.setTitle("Mrklin Control Client");
        this.primaryStage.getIcons().add(new Image("file:res/BO.png"));
        initRootLayout();
        initTabs();
    }

    /**
     * Initializes the root layout.
     */
    public void initRootLayout() {
        try {
            // Load root layout from fxml file.
            FXMLLoader loader = new FXMLLoader();
            loader.setLocation(MainApp.class.getResource("view/RootLayout.fxml"));
            rootLayout = (BorderPane) loader.load();
            rootController = loader.getController(); // only works if loader.load() has been called already
            rootController.setMainApp(this);
                      
            // Show the scene containing the root layout.
            Scene scene = new Scene(rootLayout);
            primaryStage.setScene(scene);
            primaryStage.show();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Shows the control tabs inside the root layout.
     */
    public void initTabs() {
        try {
        	FXMLLoader loader = new FXMLLoader();
            
        	loader.setLocation(MainApp.class.getResource("view/Engine.fxml"));
            enginePane = loader.load();
            engineController = loader.getController(); // only works if loader.load() has been called already
            engineController.setMainApp(this);
            
            FXMLLoader switchListLoader = new FXMLLoader();
            switchListLoader.setLocation(MainApp.class.getResource("view/SwitchList.fxml"));
            switchListPane = switchListLoader.load();
            switchListController = switchListLoader.getController();
            switchListController.setMainApp(this);
            switchListController.setSwitchControls();
            
            FXMLLoader turntableLoader = new FXMLLoader();
            turntableLoader.setLocation(MainApp.class.getResource("view/Turntable.fxml"));
            turntablePane = turntableLoader.load();
            //turntableController = turntableLoader.getController();
            //turntableController.setMainApp(this);
            
            FXMLLoader settingsLoader = new FXMLLoader();
            settingsLoader.setLocation(MainApp.class.getResource("view/Settings.fxml"));
            settingsPane = settingsLoader.load();
            settingsController = settingsLoader.getController();
            settingsController.setMainApp(this);
            settingsController.setSettings(configuration);
        	
        	// Add the tabs
        	tabs.add(new Tab("Loks", enginePane));
        	tabs.add(new Tab("Weichen", switchListPane));
        	tabs.add(new Tab("Drehscheibe", turntablePane));
        	tabs.add(new Tab("Einstellungen", settingsPane));
        	rootController.loadTabs();        	
           
        } catch (IOException e) {
            e.printStackTrace();
        }
        
    }

    /**
     * Returns the main stage.
     * @return
     */
    
    public void setStatus(String status){
    	rootController.setStatus(status);
    }
    
    public Stage getPrimaryStage() {
        return primaryStage;
    }

    public static void main(String[] args) {
        launch(args);
    }

	public void emergencyStopHandler() {
		emergencyStop();				
		for(Engine eng:engines){
			eng.setSpeed(0);
		}
		engineController.updateEngineStatus();
		setStatus("NOTAUS");
		
	}
	
	public void emergencyStop () {
		if (connectionEstablished) {
			try {
				byte[] datagram = {(byte) 0x00, (byte) Properties.SYSTEM_STOP, (byte) 0x00, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: emergencyStop");
				e.printStackTrace();
			}
		}
	}
	
	public void setEngineDirection (Engine eng) {
		if (connectionEstablished) {
			try {
				byte dir; 
				if (eng.getDirection().get())
					dir = (byte) 0x01;
				else
					dir = (byte) 0x02;
				
				byte[] datagram = {(byte) eng.getMaerklinID().get(), (byte) Properties.ENGINE_SET_DIRECTION, dir, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: engineSetDirectionRight");
				e.printStackTrace();
			}
		}
	}
	
	public void setEngineSpeed (Engine eng) {
		if (connectionEstablished) {
			try {
				int newEngineSpeed = eng.getSpeed().get();
				byte[] datagram = {(byte) eng.getMaerklinID().get(), (byte) Properties.ENGINE_SET_SPEED, (byte) ((newEngineSpeed/(1<<8))%(1<<8)), (byte) (newEngineSpeed%(1<<8))};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("engineSetSpeed");
				e.printStackTrace();
			}
		}
	}
	
	public void setSwitchState (Switch sw) {
		if (connectionEstablished) {
			try {
				byte dir; 
				if (!sw.getState().get()) 	//gerade
					dir = (byte) 0x00;
				else						//krumm
					dir = (byte) 0x01;
				
				byte[] datagram = {(byte) sw.getMaerklinID().get(), (byte) Properties.SWITCH_SET_DIRECTION, dir, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: setSwitchState");
				e.printStackTrace();
			}
		}
	}
	
	public void turntableTurnCW () {
		if (connectionEstablished) {
			try {
				byte[] datagram = {(byte) turntable.getMaerklinID().get(), (byte) Properties.SWITCH_SET_DIRECTION, (byte) 0x00, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: turntableTurnCW");
				e.printStackTrace();
			}
		}
	}
	
	public void turntableTurnCCW () {
		if (connectionEstablished) {
			try {
				byte[] datagram = {(byte) turntable.getMaerklinID().get(), (byte) Properties.SWITCH_SET_DIRECTION, (byte) 0x01, (byte) 0x00};
				out.write(datagram);
			} catch (Exception e) {
				setStatus("Error at: turntableTurnCCW");
				e.printStackTrace();
			}
		}
	}

	public void parseDatagram (byte[] datagram){
		for (Engine eng : engines){
			if(eng.getMaerklinID().get()==datagram[0]){
				if (datagram[1] == (byte) Properties.ENGINE_SET_DIRECTION)
					eng.setDirection(datagram[2]  ==(byte) 0x01);
				if (datagram[1] == (byte) Properties.ENGINE_SET_SPEED)
					eng.setSpeed(datagram[2]*(1<<8)+datagram[3]);
				if (eng.equals(engineController.getSelectedEngine()))
					engineController.updateEngineStatus();
			}
		for (Switch sw : switches){
			if(sw.getMaerklinID().get()==datagram[0]){
				if (datagram[1] == (byte) Properties.SWITCH_SET_DIRECTION){
					sw.setState(datagram[2]  == (byte) 0x01);
					sw.getController().get().updateSwitchStatus();
				}
			}
		}
		if (turntable.getMaerklinID().get()==datagram[0]){
			if (datagram[1] == (byte) Properties.SWITCH_SET_DIRECTION){
				turntable.setState(datagram[2]  ==(byte) 0x01);
				turntable.getController().get().updateSwitchStatus();
			}
		}
		}
	}
	
	
	public void establishConnetion () {
		if (connectionEstablished) {
			try {
				byte[] datagram = {(byte) 0x00, (byte) Properties.SESSION_ABORT, (byte) 0x00, (byte) 0x00};
				out.write(datagram);
				client.close();
				connectionEstablished = false;
				
			} catch (Exception e) {
				setStatus("Error at: establishConnection (shutting down connection");
				e.printStackTrace();
			}
		}
		else {
			try (Socket socket = new Socket(configuration.getSocketAddress().get().getHostName(), configuration.getSocketAddress().get().getPort())) {
				client = socket;
				client.setReuseAddress(true);
				in = client.getInputStream();
				out = client.getOutputStream();
				//client.setSoTimeout(200);
				connectionEstablished = true;
				setStatus("Verbindung zu "+client.getRemoteSocketAddress()+" aufgebaut!");
				Platform.runLater(new updateThread(this, in));
			}
			catch (Exception e) {
				setStatus("Error at: establishConnection ( " + configuration.getSocketAddress().get().toString() + " )");
				e.printStackTrace();
			}
		}
	}
	
class updateThread implements Runnable {
		
		private final MainApp CONTROLLER_INSTANCE;
		private final BufferedInputStream INPUT_STREAM;
		
		updateThread (MainApp controller, InputStream in) {
			CONTROLLER_INSTANCE = controller;
			INPUT_STREAM = new BufferedInputStream(in, Properties.OUT_BUFFER_SIZE);
		}

		@Override
		public void run() {
			int stopCounter, dataCounter, counter;
			byte buffer[] = new byte[Properties.OUT_BUFFER_SIZE];
			byte datagram[] = new byte[Properties.IN_BUFFER_SIZE];
			boolean startByte;
			try {
				stopCounter = 0;
				dataCounter = 0;
				counter = 0;
				startByte = false;
				while (connectionEstablished) {
					while (INPUT_STREAM.available() > 0 && stopCounter < 3) {
						buffer[counter%Properties.OUT_BUFFER_SIZE] = (byte) INPUT_STREAM.read();
						if (stopCounter == 1) {
							startByte = true;
						}
						if (buffer[counter%Properties.OUT_BUFFER_SIZE] == 0xff && dataCounter != 3) {
							dataCounter = 0;
							stopCounter++;
						}
						else {
							stopCounter = 0;
							dataCounter++;
							if (dataCounter == 4 && startByte) {
								for (int i = 0; i < Properties.IN_BUFFER_SIZE; i++) {
									datagram[i] = buffer[counter-Properties.IN_BUFFER_SIZE+i];
								}
								CONTROLLER_INSTANCE.parseDatagram(datagram);
							}
							else if (dataCounter > 4 && startByte) {
								throw (new Exception("Wrong data-update-format!"));
							}
						}
					}
					Thread.yield();
				}
			}
			catch (Exception e) {
				CONTROLLER_INSTANCE.setStatus("Error while updating!");
			}
		}
	}
}